"""
Calculations on notes, both in the context of octaves and without.

Vocabulary:
"semitone offset": The number of semitones away from C in the octave.
"absolute semitone offset": The number of semitones away from C0.
"semitone distance": A number of semitones.
"""

from enum import Enum
from functools import total_ordering
from typing import Self


# Enum for musical note letters
class NoteLetter(Enum):
    """
    Enum representing musical note letters.
    """

    C = 0
    D = 1
    E = 2
    F = 3
    G = 4
    A = 5
    B = 6

    def __str__(self) -> str:
        return self.name

    def __add__(self, other: int) -> "NoteLetter":
        """
        When adding letters, treat them as integers and wrap around if necessary.
        """
        new_value = (self.value + other) % len(NoteLetter)
        return NoteLetter(new_value)

    def __sub__(self, other: int) -> "NoteLetter":
        """
        When subtracting letters, treat them as integers and wrap around if necessary.
        """
        new_value = (self.value - other) % len(NoteLetter)
        return NoteLetter(new_value)

    @property
    def semitone_offset(self) -> int:
        """
        Returns how many semitones this letter is away from C.
        """
        match self:
            case NoteLetter.C:
                return 0
            case NoteLetter.D:
                return 2
            case NoteLetter.E:
                return 4
            case NoteLetter.F:
                return 5
            case NoteLetter.G:
                return 7
            case NoteLetter.A:
                return 9
            case NoteLetter.B:
                return 11

    @property
    def __int__(self) -> int:
        return self.value

    @classmethod
    def from_str(cls, s: str) -> "NoteLetter":
        """
        Converts a string to a NoteLetter by matching the string with the name of each enum variant.
        """
        for note in cls:
            if note.name == s:
                return note
        raise ValueError(f"Invalid note letter string: {s}")


class NoteAlteration(Enum):
    """
    Enum representing musical note alterations.
    """

    SHARP = 1
    FLAT = -1
    NATURAL = 0
    DOUBLE_SHARP = 2
    DOUBLE_FLAT = -2

    @property
    def semitone_difference(self) -> int:
        """
        Returns how many semitones this alteration modifies the note by.
        """
        return self.value

    def __int__(self) -> int:
        return self.semitone_difference

    def __str__(self):
        match self:
            case NoteAlteration.SHARP:
                return "♯"
            case NoteAlteration.FLAT:
                return "♭"
            case NoteAlteration.NATURAL:
                return "♮"
            case NoteAlteration.DOUBLE_SHARP:
                return "𝄪"
            case NoteAlteration.DOUBLE_FLAT:
                return "𝄫"

    @classmethod
    def from_str(cls, s: str) -> "NoteAlteration":
        match s:
            case "♯":
                return cls.SHARP
            case "#":
                return cls.SHARP
            case "♭":
                return cls.FLAT
            case "b":
                return cls.FLAT
            case "♮":
                return cls.NATURAL
            case "n":
                return cls.NATURAL
            case "𝄪":
                return cls.DOUBLE_SHARP
            case "##":
                return cls.DOUBLE_SHARP
            case "𝄫":
                return cls.DOUBLE_FLAT
            case "bb":
                return cls.DOUBLE_FLAT
            case _:
                raise ValueError(f"Invalid alteration string: {s}")


class Note:
    """
    An abstract musical note, without a specific octave.
    """

    def __init__(self, letter: NoteLetter, alteration: NoteAlteration):
        self.letter = letter
        self.alteration = alteration

    @classmethod
    def from_str(cls, s: str) -> Self:
        """
        Allows creating a Note from a string representation like "C", "C#" or "Db".
        """
        if len(s) == 1:
            return cls(NoteLetter.from_str(s), NoteAlteration.NATURAL)

        letter = NoteLetter.from_str(s[0])
        alteration = NoteAlteration.from_str(s[1:])
        return cls(letter, alteration)

    @property
    def semitone_offset(self) -> int:
        """
        Returns how many semitones away from C this note is. The semitone offset is the sum of the letter's semitone offset and the alteration's semitone difference.

        Examples:
        - C = 0
        - C# = 1
        - C## = 2
        - Cb = -1
        - Cbb = -2
        - B# = 12
        - B## = 13
        """

        return self.letter.semitone_offset + self.alteration.semitone_difference

    @classmethod
    def from_semitone_offset(cls, semitone_offset: int) -> set[Self]:
        """
        Returns all possible notes that are a specific number of semitones away from C.

        Examples:
        - 0 -> C and Dbb
        - -1 -> Cb
        - -2 -> Cbb
        - 1 -> C# and Db
        - 7 -> G, F## and Abb
        - 8 -> G# and Ab
        - 11 -> B and A##
        - 12 -> B#
        - 13 -> B##
        """

        # Loop through all possible enumerations of NoteLetter and NoteAlteration
        notes = set(
            cls(note_letter, note_alteration)
            for note_letter in NoteLetter
            for note_alteration in NoteAlteration
            if note_letter.semitone_offset + note_alteration.semitone_difference
            == semitone_offset
        )

        return notes

    def __eq__(self, other: object) -> bool:
        """
        Two non-octaved notes are only equal if they have the same letter and alteration.
        """
        if not isinstance(other, Note):
            return NotImplemented
        return self.letter == other.letter and self.alteration == other.alteration

    # lt is not implemented because it doesn't make sense to compare notes without an octave

    def __str__(self):
        return (
            f"{self.letter}{self.alteration}"
            if self.alteration != NoteAlteration.NATURAL
            else str(self.letter)
        )

    def __repr__(self):
        return f"Note({self.letter}, {self.alteration})"

    def __hash__(self) -> int:
        return hash((self.letter, self.alteration))


class Interval(Enum):
    """
    Enum representing musical intervals with their semitone and letter distances.
    """

    PERFECT_UNISON = (0, 0)
    MINOR_SECOND = (1, 1)
    MAJOR_SECOND = (2, 1)
    MINOR_THIRD = (3, 2)  # 3 semitones, 2 letter steps
    MAJOR_THIRD = (4, 2)  # etc.
    PERFECT_FOURTH = (5, 3)
    PERFECT_FIFTH = (7, 4)
    MINOR_SIXTH = (8, 5)
    MAJOR_SIXTH = (9, 5)
    MINOR_SEVENTH = (10, 6)
    MAJOR_SEVENTH = (11, 6)
    PERFECT_OCTAVE = (12, 7)

    @property
    def semitone_distance(self) -> int:
        """
        Get the semitone distance of the interval.
        """
        return self.value[0]

    @property
    def letter_distance(self) -> int:
        """
        Get the letter distance of the interval.
        """
        return self.value[1]


@total_ordering
class NoteInOctave:
    """
    A concrete musical note in a specific octave
    """

    def __init__(self, note: Note, octave: int):
        self.note = note
        assert octave >= 0, "Octaves cannot be negative"
        self.octave = octave

    @classmethod
    def from_absolute_semitone_offset(cls, absolute_semitone_offset: int) -> set[Self]:
        """
        Returns all possible notes that are a specific number of semitones away from C0.
        """

        # Find the candidate offset within the octave, and the candidate octave.
        #
        # Edge cases close to octave changes need to be handled carefully. Let X denote an octave.
        #
        # Case 1: "CX" can be interpreted as being in octave X-1 (B#)
        # Case 2: "C#X" can be interpreted as being in octave X-1 (B##)
        # Case 3: "BY" can be interpreted as being in octave Y+1 (Cb)
        # Case 4: "BbY" can be interpreted as being in octave Y+1 (Cbb)

        if absolute_semitone_offset % 12 == 0:
            semitone_offset_candidates = {0, 12}
            candidate_octaves = {
                absolute_semitone_offset // 12,
                absolute_semitone_offset // 12 - 1,
            }
        elif absolute_semitone_offset % 12 == 1:
            semitone_offset_candidates = {1, 13}
            candidate_octaves = {
                absolute_semitone_offset // 12,
                absolute_semitone_offset // 12 - 1,
            }
        elif absolute_semitone_offset % 12 == 11:
            semitone_offset_candidates = {11, -1}
            candidate_octaves = {
                absolute_semitone_offset // 12,
                absolute_semitone_offset // 12 + 1,
            }
        elif absolute_semitone_offset % 12 == 10:
            semitone_offset_candidates = {10, -2}
            candidate_octaves = {
                absolute_semitone_offset // 12,
                absolute_semitone_offset // 12 + 1,
            }
        else:
            semitone_offset_candidates = {absolute_semitone_offset % 12}
            candidate_octaves = {absolute_semitone_offset // 12}

        # Find all possible non-octaved notes
        notes = set.union(
            *(
                Note.from_semitone_offset(semitone_offset)
                for semitone_offset in semitone_offset_candidates
            )
        )

        # Keep only notes and octaves such that their semitone offset adds up to the absolute semitone offset
        return set(
            cls(note, octave)
            for note in notes
            for octave in candidate_octaves
            if note.semitone_offset + octave * 12 == absolute_semitone_offset
        )

    def from_semitone_distance(self, semitone_offset: int) -> set["NoteInOctave"]:
        """
        Returns all possible notes that are a specific number of semitones away from this note
        """

        return NoteInOctave.from_absolute_semitone_offset(
            self.absolute_semitone_offset + semitone_offset
        )

    def __add__(self, interval: Interval) -> "NoteInOctave":
        """
        Adds an interval to the note, potentially changing its octave.
        """

        # This is only implemented for notes without double accidentals. Otherwise we would have to handle
        # complex cases like B## + M3 = D###4
        if self.note.alteration in (
            NoteAlteration.DOUBLE_SHARP,
            NoteAlteration.DOUBLE_FLAT,
        ):
            raise ValueError("Cannot add intervals to notes with double accidentals.")

        # Find all possible notes that are a specific number of semitones away from this note
        feasible_notes = self.from_semitone_distance(interval.semitone_distance)

        # Only one of these notes will have the correct letter
        new_notes = [
            note
            for note in feasible_notes
            if note.letter - interval.letter_distance == self.letter
        ]
        assert len(new_notes) == 1, (
            "There should only be one correct note after adding an interval"
        )
        new_note = new_notes[0]

        return new_note

    def __str__(self):
        return f"{self.note}{self.octave}"

    @property
    def absolute_semitone_offset(self) -> int:
        """
        The absolute semitone offset of the note is the semitone offset from C0.
        """
        # absolute_semitone_offset = 12*self.octave + self.note.letter.semitone_offset + self.note.alteration.semitone_difference
        absolute_semitone_offset = 12 * self.octave + self.note.semitone_offset

        return absolute_semitone_offset

    def __int__(self):
        """
        The integer representation of the note is absolute semitone offset.
        """
        return self.absolute_semitone_offset

    def __repr__(self):
        return f"NoteInOctave({self.note}, {self.octave})"

    def __eq__(self, other: object) -> bool:
        """
        Comparison is forwarded to the absolute semitone offset.
        """
        assert isinstance(other, NoteInOctave), (
            "Cannot compare NoteInOctave with non-NoteInOctave object"
        )
        return self.absolute_semitone_offset == other.absolute_semitone_offset

    def __lt__(self, other: object) -> bool:
        """
        Comparison is forwarded to the absolute semitone offset.
        """
        assert isinstance(other, NoteInOctave), (
            "Cannot compare NoteInOctave with non-NoteInOctave object"
        )
        return self.absolute_semitone_offset < other.absolute_semitone_offset

    def __hash__(self) -> int:
        return hash((self.note, self.octave))

    @classmethod
    def from_str(cls, s: str) -> Self:
        """
        Converts a string to an octaved note by splitting the string into note and octave parts.

        Examples: NoteInOctave.from_str("C4"), NoteInOctave.from_str("D♯5"), NoteInOctave.from_str("B♭3")
        """
        note_str = s[:-1]
        octave_str = s[-1]
        note = Note.from_str(note_str)
        octave = int(octave_str)
        return cls(note, octave)

    # Forward some properties to the underlying note to avoid breaking law of demeter
    @property
    def letter(self) -> NoteLetter:
        return self.note.letter

    @property
    def alteration(self) -> NoteAlteration:
        return self.note.alteration


class Chord:
    """
    A Chord is a set of octaved notes.
    """

    def __init__(self, notes: set[NoteInOctave]):
        self.notes = notes

    def __iter__(self):
        return iter(self.notes)

    def __repr__(self):
        return f"Chord({{{','.join(str(note) for note in sorted(self.notes))}}})"

    def __str__(self):
        return f"{{{','.join(str(note) for note in sorted(self.notes))}}}"



from pymusictheory import Chord, Interval, Note, NoteInOctave, NoteLetter


class TestNoteLetter:
    def test_addition(self):
        actual = [
            NoteLetter.C + 4,
            NoteLetter.E + 7,
            NoteLetter.A + 2,
        ]

        expected = [
            NoteLetter.G,
            NoteLetter.E,
            NoteLetter.C,
        ]

        assert actual == expected, f"Expected {expected}, but got {actual}"

    def test_subtraction(self):
        actual = [
            NoteLetter.C - 4,
            NoteLetter.E - 7,
            NoteLetter.A - 2,
        ]

        expected = [
            NoteLetter.F,
            NoteLetter.E,
            NoteLetter.F,
        ]

        assert actual == expected, f"Expected {expected}, but got {actual}"


class TestNote:
    def test_semitone_offset(self):
        assert Note.from_str("C").semitone_offset == 0
        assert Note.from_str("C#").semitone_offset == 1
        assert Note.from_str("C##").semitone_offset == 2
        assert Note.from_str("Cb").semitone_offset == -1
        assert Note.from_str("Cbb").semitone_offset == -2
        assert Note.from_str("Eb").semitone_offset == 3
        assert Note.from_str("F#").semitone_offset == 6
        assert Note.from_str("B#").semitone_offset == 12
        assert Note.from_str("B##").semitone_offset == 13

    def test_from_semitone_offset(self):
        assert Note.from_semitone_offset(0) == {
            Note.from_str("C"),
            Note.from_str("Dbb"),
        }
        assert Note.from_semitone_offset(-1) == {Note.from_str("Cb")}
        assert Note.from_semitone_offset(-2) == {Note.from_str("Cbb")}
        assert Note.from_semitone_offset(1) == {
            Note.from_str("C#"),
            Note.from_str("Db"),
        }
        assert Note.from_semitone_offset(7) == {
            Note.from_str("G"),
            Note.from_str("F##"),
            Note.from_str("Abb"),
        }
        assert Note.from_semitone_offset(8) == {
            Note.from_str("G#"),
            Note.from_str("Ab"),
        }
        assert Note.from_semitone_offset(11) == {
            Note.from_str("B"),
            Note.from_str("A##"),
        }
        assert Note.from_semitone_offset(12) == {Note.from_str("B#")}
        assert Note.from_semitone_offset(13) == {Note.from_str("B##")}

    def test_eq(self):
        # Should be equal
        assert Note.from_str("C") == Note.from_str("C")
        assert Note.from_str("Cb") == Note.from_str("Cb")
        assert Note.from_str("G#") == Note.from_str("G#")
        # Should not be equal
        assert Note.from_str("C") != Note.from_str("D")
        assert Note.from_str("Cb") != Note.from_str("B#")
        assert Note.from_str("G#") != Note.from_str("Ab")
        assert Note.from_str("A##") != Note.from_str("B")
        assert Note.from_str("Bbb") != Note.from_str("A")


class TestNoteInOctave:
    def test_eq(self):
        # Should be equal
        assert NoteInOctave.from_str("C4") == NoteInOctave.from_str("C4")
        assert NoteInOctave.from_str("Cb4") == NoteInOctave.from_str("B3")
        assert NoteInOctave.from_str("C4") == NoteInOctave.from_str("B#3")
        assert NoteInOctave.from_str("E#4") == NoteInOctave.from_str("F4")
        assert NoteInOctave.from_str("G#4") == NoteInOctave.from_str("Ab4")
        # Should not be equal
        assert NoteInOctave.from_str("C4") != NoteInOctave.from_str("D4")
        assert NoteInOctave.from_str("C4") != NoteInOctave.from_str("Cb5")

    def test_absolute_semitone_offset(self):
        assert NoteInOctave.from_str("C1").absolute_semitone_offset == 12
        assert NoteInOctave.from_str("C2").absolute_semitone_offset == 24
        assert NoteInOctave.from_str("Cb2").absolute_semitone_offset == 23
        assert NoteInOctave.from_str("B#1").absolute_semitone_offset == 24
        assert NoteInOctave.from_str("G#2").absolute_semitone_offset == 32

    def test_from_absolute_semitone_offset(self):
        assert NoteInOctave.from_absolute_semitone_offset(28) == set(
            [
                NoteInOctave.from_str("E2"),
                NoteInOctave.from_str("Fb2"),
                NoteInOctave.from_str("D##2"),
            ]
        )
        assert NoteInOctave.from_absolute_semitone_offset(31) == set(
            [
                NoteInOctave.from_str("G2"),
                NoteInOctave.from_str("F##2"),
                NoteInOctave.from_str("Abb2"),
            ]
        )
        assert NoteInOctave.from_absolute_semitone_offset(32) == set(
            [
                NoteInOctave.from_str("G#2"),
                NoteInOctave.from_str("Ab2"),
            ]
        )
        # Four edge cases
        assert NoteInOctave.from_absolute_semitone_offset(12) == set(
            [
                NoteInOctave.from_str("C1"),
                NoteInOctave.from_str("B#0"),
                NoteInOctave.from_str("Dbb1"),
            ]
        )
        assert NoteInOctave.from_absolute_semitone_offset(13) == set(
            [
                NoteInOctave.from_str("C#1"),
                NoteInOctave.from_str("B##0"),
                NoteInOctave.from_str("Db1"),
            ]
        )
        assert NoteInOctave.from_absolute_semitone_offset(11) == set(
            [
                NoteInOctave.from_str("Cb1"),
                NoteInOctave.from_str("B0"),
                NoteInOctave.from_str("A##0"),
            ]
        )
        assert NoteInOctave.from_absolute_semitone_offset(10) == set(
            [
                NoteInOctave.from_str("Cbb1"),
                NoteInOctave.from_str("Bb0"),
                NoteInOctave.from_str("A#0"),
            ]
        )

    def test_from_semitone_distance(self):
        actual = NoteInOctave.from_str("C4").from_semitone_distance(0)
        expected = {
            NoteInOctave.from_str("C4"),
            NoteInOctave.from_str("B#3"),
            NoteInOctave.from_str("Dbb4"),
        }
        assert actual == expected

        actual = NoteInOctave.from_str("Cb4").from_semitone_distance(0)
        expected = {
            NoteInOctave.from_str("Cb4"),
            NoteInOctave.from_str("B3"),
            NoteInOctave.from_str("A##3"),
        }
        assert actual == expected

        actual = NoteInOctave.from_str("B#4").from_semitone_distance(0)
        expected = {
            NoteInOctave.from_str("B#4"),
            NoteInOctave.from_str("C5"),
            NoteInOctave.from_str("Dbb5"),
        }
        assert actual == expected

        actual = NoteInOctave.from_str("G4").from_semitone_distance(5)
        expected = {
            NoteInOctave.from_str("C5"),
            NoteInOctave.from_str("B#4"),
            NoteInOctave.from_str("Dbb5"),
        }
        assert actual == expected

        actual = NoteInOctave.from_str("A3").from_semitone_distance(5)
        expected = {
            NoteInOctave.from_str("D4"),
            NoteInOctave.from_str("C##4"),
            NoteInOctave.from_str("Ebb4"),
        }
        assert actual == expected

        actual = NoteInOctave.from_str("A3").from_semitone_distance(6)
        expected = {
            NoteInOctave.from_str("D#4"),
            NoteInOctave.from_str("Eb4"),
            NoteInOctave.from_str("Fbb4"),
        }
        assert actual == expected

    def test_add_interval(self):
        """
        Test that adding intervals works as expected.

        Tested so far (in octave 4):
        - PERFECT_UNISON
        - MINOR_THIRD
        - MAJOR_THIRD
        - PERFECT_FIFTH
        - PERFECT_OCTAVE

        Not yet tested:
        - MINOR_SECOND
        - MAJOR_SECOND
        - PERFECT_FOURTH
        - MINOR_SIXTH
        - MAJOR_SIXTH
        - MINOR_SEVENTH
        - MAJOR_SEVENTH

        Also tested a perfect fifth with a different octave.
        """

        # Random perfect unisons
        assert NoteInOctave.from_str(
            "C4"
        ) + Interval.PERFECT_UNISON == NoteInOctave.from_str("C4")
        assert NoteInOctave.from_str(
            "C#4"
        ) + Interval.PERFECT_UNISON == NoteInOctave.from_str("C#4")
        assert NoteInOctave.from_str(
            "D4"
        ) + Interval.PERFECT_UNISON == NoteInOctave.from_str("D4")

        # Random minor thirds
        assert NoteInOctave.from_str(
            "C4"
        ) + Interval.MINOR_THIRD == NoteInOctave.from_str("Eb4")
        assert NoteInOctave.from_str(
            "C#4"
        ) + Interval.MINOR_THIRD == NoteInOctave.from_str("E4")
        assert NoteInOctave.from_str(
            "D4"
        ) + Interval.MINOR_THIRD == NoteInOctave.from_str("F4")

        # There are no minor thirds resulting in double sharps, if we only consider starting notes with single accidentals

        # Minor thirds resulting in double flats
        assert NoteInOctave.from_str(
            "Fb4"
        ) + Interval.MINOR_THIRD == NoteInOctave.from_str("Abb4")
        assert NoteInOctave.from_str(
            "Cb4"
        ) + Interval.MINOR_THIRD == NoteInOctave.from_str("Ebb4")

        # Random major thirds
        assert NoteInOctave.from_str(
            "C4"
        ) + Interval.MAJOR_THIRD == NoteInOctave.from_str("E4")
        assert NoteInOctave.from_str(
            "C#4"
        ) + Interval.MAJOR_THIRD == NoteInOctave.from_str("E#4")
        assert NoteInOctave.from_str(
            "D4"
        ) + Interval.MAJOR_THIRD == NoteInOctave.from_str("F#4")

        # Major thirds resulting in double sharps
        assert NoteInOctave.from_str(
            "B#4"
        ) + Interval.MAJOR_THIRD == NoteInOctave.from_str("D##5")
        assert NoteInOctave.from_str(
            "D#4"
        ) + Interval.MAJOR_THIRD == NoteInOctave.from_str("F##4")
        assert NoteInOctave.from_str(
            "E#4"
        ) + Interval.MAJOR_THIRD == NoteInOctave.from_str("G##4")
        assert NoteInOctave.from_str(
            "A#4"
        ) + Interval.MAJOR_THIRD == NoteInOctave.from_str("C##5")

        # There are no major thirds resulting in double flats, if we only consider starting notes with single accidentals

        # Random perfect fifths
        assert NoteInOctave.from_str(
            "C4"
        ) + Interval.PERFECT_FIFTH == NoteInOctave.from_str("G4")
        assert NoteInOctave.from_str(
            "C#4"
        ) + Interval.PERFECT_FIFTH == NoteInOctave.from_str("G#4")
        assert NoteInOctave.from_str(
            "D4"
        ) + Interval.PERFECT_FIFTH == NoteInOctave.from_str("A4")

        # Random perfect unisons
        assert NoteInOctave.from_str(
            "C4"
        ) + Interval.PERFECT_OCTAVE == NoteInOctave.from_str("C5")
        assert NoteInOctave.from_str(
            "C#4"
        ) + Interval.PERFECT_OCTAVE == NoteInOctave.from_str("C#5")
        assert NoteInOctave.from_str(
            "D4"
        ) + Interval.PERFECT_OCTAVE == NoteInOctave.from_str("D5")

        # Perfect fifths in different octaves
        assert NoteInOctave.from_str(
            "G4"
        ) + Interval.PERFECT_FIFTH == NoteInOctave.from_str("D5")
        assert NoteInOctave.from_str(
            "G#4"
        ) + Interval.PERFECT_FIFTH == NoteInOctave.from_str("D#5")
        assert NoteInOctave.from_str(
            "Ab4"
        ) + Interval.PERFECT_FIFTH == NoteInOctave.from_str("Eb5")

        # Perfect fifths with double sharps
        assert NoteInOctave.from_str(
            "B#4"
        ) + Interval.PERFECT_FIFTH == NoteInOctave.from_str("F##5")
        assert NoteInOctave.from_str(
            "E#4"
        ) + Interval.PERFECT_FIFTH == NoteInOctave.from_str("B#4")

        # There are no perfect fifths resulting in double flats, if we only consider starting notes with single accidentals

        # We expect an error if we try to add an interval to a note that already has a double accidental
        # This would necessitate handling of complex cases like B## + M3 = D### which are
        # never used in practice

        # Assert that code should raise an error
        try:
            NoteInOctave.from_str("B##4") + Interval.MAJOR_THIRD
        except ValueError as e:
            assert str(e) == "Cannot add intervals to notes with double accidentals."
        else:
            assert False, "Expected ValueError, but no error was raised"


class TestChord:
    def test_repr(self):
        # Test the representation of a chord. It should be sorted
        chord = Chord(
            {
                NoteInOctave.from_str("C4"),
                NoteInOctave.from_str("G4"),
                NoteInOctave.from_str("E4"),
            }
        )
        assert repr(chord) == r"Chord({C4,E4,G4})"

    def test_str(self):
        # Test the string representation of a chord. It should be sorted
        chord = Chord(
            {
                NoteInOctave.from_str("C4"),
                NoteInOctave.from_str("G4"),
                NoteInOctave.from_str("E4"),
            }
        )
        assert str(chord) == r"{C4,E4,G4}"
